/* $Id: tess_heap.h,v 1.4 1999/09/17 06:35:26 gareth Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * $Log: tess_heap.h,v $
 * Revision 1.4  1999/09/17 06:35:26  gareth
 * Winding rule updates.
 *
 * Revision 1.3  1999/09/14 22:47:21  gareth
 * Added heap initialization function.
 *
 * Revision 1.2  1999/09/13 22:20:13  gareth
 * Fixed file headers.  Tracking down macro bugs.
 *
 */

/*****************************************************************************
 *
 * GLU 1.3 Polygon Tessellation - Implementation of priority queue
 *
 * Gareth Hughes <garethh@lucent.com>, April 1999
 *
 *****************************************************************************/

#ifndef __GLU_TESS_HEAP_H__
#define __GLU_TESS_HEAP_H__

#include "tess_typedefs.h"

#ifdef __cplusplus
extern "C" {
#endif

#define HEAP_ALLOC	8

#define HEAP_VERTEX(v)		((tess_vertex_t *) v)

typedef struct heap_elt_s
{
    GLuint		index;
    GLdouble		value;
    void		*ptr;
    struct heap_elt_s	*previous, *next;
} heap_elt_t;

typedef struct heap_s
{
    GLuint		length;
    GLuint		size;
    GLuint		flags;
    heap_elt_t		**elements;
} heap_t;

heap_t		*heap_init();
void		heap_build( heap_t *heap );
GLboolean	heap_insert( heap_t *heap, heap_elt_t *element );
heap_elt_t	*heap_extract_max( heap_t *heap );
heap_elt_t	*heap_delete( heap_t *heap, GLuint n );

void		heap_cleanup( heap_t *heap );

#ifdef __cplusplus
}
#endif

#endif /* __GLU_TESS_HEAP_H__ */
