/* $Id: tess_macros.h,v 1.4 1999/09/17 06:35:59 gareth Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 *
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * $Log: tess_macros.h,v $
 * Revision 1.4  1999/09/17 06:35:59  gareth
 * Winding rule updates.
 *
 * Revision 1.3  1999/09/13 22:20:13  gareth
 * Fixed file headers.  Tracking down macro bugs.
 *
 */

/*****************************************************************************
 *
 * GLU 1.3 Polygon Tessellation by Gareth Hughes <garethh@lucent.com>
 *
 *****************************************************************************/

#ifndef __GLU_TESS_MACROS_H__
#define __GLU_TESS_MACROS_H__

#include "../src/macros.h"
#include "../src/mmath.h"

#define X	0
#define Y	1
#define Z	2

/* Limits: */
#ifndef DBL_MAX
#define DBL_MAX		1E+37
#endif

#define MAX_GLDOUBLE	DBL_MAX
#define EQUAL_EPSILON	1E-09


/* Comparison: */

#define COMPARE_3DV( A, B )				\
( ( ABSD( (A)[X] - (B)[X] ) < EQUAL_EPSILON ) &&	\
  ( ABSD( (A)[Y] - (B)[Y] ) < EQUAL_EPSILON ) &&	\
  ( ABSD( (A)[Z] - (B)[Z] ) < EQUAL_EPSILON ) )


/* Length: */

#define LEN_3DV( V )		LEN_3FV( V )

#define LEN_SCALAR( V1, V2 )	( GL_SQRT( (V1)*(V1) + (V2)*(V2) ) )

#define NORMALIZE_3DV( V )	NORMALIZE_3FV( V )


/* Bounding boxes: */

#define CLEAR_BBOX_2DV( MINS, MAXS )		\
do {						\
    (MINS)[X] = (MINS)[Y] = MAX_GLDOUBLE;	\
    (MAXS)[X] = (MAXS)[Y] = -MAX_GLDOUBLE;	\
} while ( 0 )

#define ACC_BBOX_2V( V, MINS, MAXS )		\
do {						\
    if ( (V)[X] < (MINS)[X] ) {			\
	(MINS)[X] = (V)[X];			\
    }						\
    if ( (V)[X] > (MAXS)[X] ) {			\
	(MAXS)[X] = (V)[X];			\
    }						\
    if ( (V)[Y] < (MINS)[Y] ) {			\
	(MINS)[Y] = (V)[Y];			\
    }						\
    if ( (V)[Y] > (MAXS)[Y] ) {			\
	(MAXS)[Y] = (V)[Y];			\
    }						\
} while ( 0 )

/* Test if bbox B is inside bbox A: */
#define CONTAINS_BBOX_2V( AMINS, AMAXS, BMINS, BMAXS )			\
( ( (AMINS)[X] < (BMINS)[X] ) && ( (AMINS)[Y] < (BMINS)[Y] ) &&		\
  ( (AMAXS)[X] > (BMAXS)[X] ) && ( (AMAXS)[Y] > (BMAXS)[Y] ) )

#define INSIDE_BBOX_2V( V, MINS, MAXS )			\
( ( (MINS)[X] < (V)[X] ) && ( (MINS)[Y] < (V)[Y] ) &&	\
  ( (MAXS)[X] > (V)[X] ) && ( (MAXS)[Y] > (V)[Y] ) )

#endif /* __GLU_TESS_MACROS_H__ */
