/* $Id: tess_typedefs.h,v 1.5 1999/09/17 06:39:28 gareth Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * $Log: tess_typedefs.h,v $
 * Revision 1.5  1999/09/17 06:39:28  gareth
 * Winding rule updates.
 *
 * Revision 1.4  1999/09/16 06:45:50  gareth
 * Misc winding rule bug fixes.
 *
 * Revision 1.3  1999/09/13 22:20:13  gareth
 * Fixed file headers.  Tracking down macro bugs.
 *
 */

/*****************************************************************************
 *
 * GLU 1.3 Polygon Tessellation by Gareth Hughes <garethh@lucent.com>
 *
 *****************************************************************************/

#ifndef __GLU_TESS_TYPEDEFS_H__
#define __GLU_TESS_TYPEDEFS_H__

#include <stdarg.h>
#include <stdio.h>

#include "gluP.h"

#include "tess_hash.h"

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************************
 * Tessellation type definitions:
 *****************************************************************************/
typedef struct
{
    void (GLCALLBACK *begin)( GLenum );
    void (GLCALLBACK *beginData)( GLenum , void * );
    void (GLCALLBACK *edgeFlag)( GLboolean );
    void (GLCALLBACK *edgeFlagData)( GLboolean , void * );
    void (GLCALLBACK *vertex)( void * );
    void (GLCALLBACK *vertexData)( void *, void * );
    void (GLCALLBACK *end)( void );
    void (GLCALLBACK *endData)( void * );
    void (GLCALLBACK *error)( GLenum );
    void (GLCALLBACK *errorData)( GLenum , void * );
    void (GLCALLBACK *combine)( GLdouble[3], void *[4],
				GLfloat[4], void ** );
    void (GLCALLBACK *combineData)( GLdouble[3], void *[4],
				    GLfloat[4], void **,
				    void * );
} tess_callbacks_t;

typedef struct
{
    GLdouble		normal[3];
    GLdouble		dist;
} tess_plane_t;

typedef struct vertex_s
{
    GLuint		index;
    void		*data;
    GLdouble		coords[3];
    GLdouble		v[2];
    GLboolean		edge_flag;
    GLdouble		angle;
    GLuint		label;
    struct vertex_s	*shadow_vertex;
    struct vertex_s	*next, *previous;
} tess_vertex_t;

typedef struct contour_s
{
    GLenum		type;
    tess_plane_t	plane;
    GLdouble		area;
    GLenum		orientation;
    GLuint		label;
    GLdouble		rotx, roty;
    GLdouble		mins[2], maxs[2];
    GLuint		vertex_count;
    tess_vertex_t	*vertices, *last_vertex;
    hashtable_t		*reflex_vertices;
    hashtable_t		*cross_vertices;
    struct contour_s	*next, *previous;
} tess_contour_t;


/*****************************************************************************
 * Labels for edge/contour intersections:
 *****************************************************************************/
#define INSIDE		1
#define OUTSIDE		2
#define SHARED1		3
#define SHARED2		4
#define ISECTED		5

#ifdef __cplusplus
}
#endif

#endif /* __GLU_TESS_TYPEDEFS_H__ */
