/* $Id: vbxform.c,v 1.7 1999/09/20 14:30:22 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */



#ifdef PC_HEADER
#include "all.h"
#else
#include <stdlib.h>
#include <stdio.h>
#include "context.h"
#include "cva.h"
#include "clip.h"
#include "eval.h"
#include "dlist.h"
#include "fog.h"
#include "light.h"
#include "macros.h"
#include "matrix.h"
#include "mmath.h"
#include "pipeline.h"
#include "shade.h"
#include "texture.h"
#include "types.h"
#include "varray.h"
#include "vb.h"
#include "vbcull.h"
#include "vbfill.h"
#include "vbrender.h"
#include "vbxform.h"
#include "xform.h"
#ifdef XFree86Server
#include "GL/xf86glx.h"
#endif
#endif


void gl_maybe_transform_vb( struct immediate *IM )
{
   GLcontext *ctx = IM->backref;

   if (ctx->NewState)
      gl_update_state(ctx);

   gl_compute_orflag( IM );

   if (ctx->ExecuteFlag) 
      gl_cva_compile_cassette( ctx, IM );
   else
      gl_fixup_input(ctx, IM);
   
   if (ctx->CompileFlag) 
      gl_compile_cassette( ctx );
      
   gl_reset_input( ctx );
}


void gl_flush_vb( GLcontext *ctx, const char *where )
{
   struct immediate *IM = ctx->input;

   if (MESA_VERBOSE&VERBOSE_PIPELINE)
      fprintf(stderr, "gl_flush_vb: %s\n", where);

   gl_maybe_transform_vb( IM );
}



#define CLIP_ALL_BITS    0x3f
#define RESET_VEC(v, t, s) (v.start = t v.data[s], v.count = 0)  

/* 
 *
 */
void gl_reset_vb( struct vertex_buffer *VB )
{
   GLuint copy;
   GLuint dst;
   GLuint start;
   struct immediate *IM = VB->IM;
   GLubyte clipor = VB->ClipOrMask;

   if (!VB->CullDone)
      gl_fast_copy_vb( VB );
      
   copy = VB->CopyCount;
   start = 3-copy;

   VB->CopyStart = start;
   VB->ClipOrMask = 0;
   VB->ClipAndMask = CLIP_ALL_BITS;

   if (IM) 
   {
      RESET_VEC(IM->v.Obj, (GLfloat *), VB_START);
      RESET_VEC(IM->v.Normal, (GLfloat *), VB_START);
      RESET_VEC(IM->v.TexCoord[0], (GLfloat *), VB_START);
      RESET_VEC(IM->v.TexCoord[1], (GLfloat *), VB_START);
      RESET_VEC(IM->v.Index, &, VB_START);
      RESET_VEC(IM->v.Elt, &, VB_START);
      RESET_VEC(IM->v.EdgeFlag, &, VB_START);
      RESET_VEC(IM->v.Color, (GLubyte *), VB_START);
      RESET_VEC(VB->Clip, (GLfloat *), VB_START);
      RESET_VEC(VB->Eye, (GLfloat *), VB_START);
      RESET_VEC(VB->Win, (GLfloat *), VB_START);
      RESET_VEC(VB->BColor, (GLubyte *), VB_START); 
      RESET_VEC(VB->BIndex, &, VB_START);
      
      for (dst = start ; dst < VB_START ; dst++) 
      {
	 GLuint src = VB->Copy[dst];
 
	 if (0)
	    fprintf(stderr, "copying vertex %u to %u\n", src, dst);

	 COPY_4FV( VB->Clip.data[dst], VEC_ELT(VB->ClipPtr, GLfloat, src) ); 
	 COPY_4FV( VB->Win.data[dst], VB->Win.data[src] ); 


	 VB->UserClipMask[dst] = VB->UserClipMask[src];
	 VB->ClipMask[dst] = VB->ClipMask[src] & ~CLIP_CULLED_BIT;   
	 VB->ClipAndMask &= VB->ClipMask[dst];
	 VB->ClipOrMask  |= VB->ClipMask[dst];
	 VB->ClipMask[src] = 0;	/* hack for bounds_cull_vb */

	 COPY_4UBV( IM->Color[dst], IM->Color[src] );
	 COPY_4UBV( VB->Spec[0][dst], VB->Spec[0][src] );
	 COPY_4UBV( VB->Spec[1][dst], VB->Spec[1][src] );
	 COPY_4UBV( VB->BColor.data[dst], VB->BColor.data[src] );
	 IM->Index[dst] = IM->Index[src];
	 VB->BIndex.data[dst] = VB->BIndex.data[src];

	 if (VB->TexCoordPtr[0] == &IM->v.TexCoord[0])
	    COPY_4FV( IM->TexCoord[0][dst], IM->TexCoord[0][src] );
	    
	 if (VB->TexCoordPtr[1] == &IM->v.TexCoord[1])
	    COPY_4FV( IM->TexCoord[1][dst], IM->TexCoord[1][src] );

	 IM->Elt[dst] = IM->Elt[src];
	 VB->SavedOrFlag |= IM->Flag[src];
      } 
      
      VB->CullDone = 0;
   }

   if (clipor & CLIP_USER_BIT) 
      MEMSET(VB->UserClipMask + VB->Start, 0, VB->Count);

   VB->NormCullStart = 0;
   VB->Parity = (VB->LastPrimitive^VB->Count)&1;
   VB->LastPrimitive = VB_START; 
   VB->PurgeFlags = 0;
   VB->EarlyCull = 1;
   VB->Culled = 0;
   VB->BoundsPtr = 0;
   VB->NormalLengthPtr = 0;  
   VB->Indirect = 0;
   VB->LastCount = VB->Count;
   VB->Culled = 0;
}




/* Copy the untransformed parts of the overlapping vertices from one
 * immediate struct to another (or possibly the same one).  
 *
 * Only copy those elements which are genuinely untransformed.  Others
 * are done in gl_reset_vb.
 */
void gl_copy_prev_vertices( struct vertex_buffer *VB,
			    struct immediate *prev,
			    struct immediate *next )
{
   GLuint dst;
   GLuint flags = VB->pipeline->inputs;

   for (dst = VB->CopyStart ; dst < VB_START ; dst++) 
   {
      GLuint src = VB->Copy[dst];

      COPY_4FV( next->Obj[dst], prev->Obj[src] );

      if ((flags&VERT_TEX0_ANY) && VB->TexCoordPtr[0] == &prev->v.TexCoord[0])
	 COPY_4FV( next->TexCoord[0][dst], prev->TexCoord[0][src] );
	    
      if ((flags&VERT_TEX1_ANY) && VB->TexCoordPtr[1] == &prev->v.TexCoord[1])
	 COPY_4FV( next->TexCoord[1][dst], prev->TexCoord[1][src] );

      COPY_4UBV( next->Color[dst], prev->Color[src] );
      next->Index[dst] = prev->Index[src];
      next->EdgeFlag[dst] = prev->EdgeFlag[src];

      next->Elt[dst] = prev->Elt[src];
      VB->SavedOrFlag |= prev->Flag[src];
   }
}



void RESET_IMMEDIATE( GLcontext *ctx ) 
{
   if (ctx->VB->prev_buffer != ctx->VB->IM) {
      /* Should only get here if we are trying to use the internal
       * interfaces, eg gl_Vertex3f(), gl_Begin() from inside a
       * display list.  In this case, it is necessary to pull the
       * current values into the ctx->VB.store buffer, because this
       * may not have been done.
       */
      FLUSH_VB( ctx, "RESET_IMMEDIATE" );
      gl_reset_input( ctx );
   }
}


/* Called to initialize new buffers, and to recycle old ones.
 */
void gl_reset_input( GLcontext *ctx )
{
   struct immediate *IM = ctx->input;

   MEMSET(IM->Flag, 0, sizeof(GLuint) * (IM->Count+2));

   RESET_VEC(IM->v.Obj, (GLfloat *), VB_START);
   RESET_VEC(IM->v.Normal, (GLfloat *), VB_START);
   RESET_VEC(IM->v.Color, (GLubyte *), VB_START);
   RESET_VEC(IM->v.Index, &, VB_START);
   RESET_VEC(IM->v.Elt, &, VB_START);
   RESET_VEC(IM->v.TexCoord[0], (GLfloat *), VB_START);
   RESET_VEC(IM->v.TexCoord[1], (GLfloat *), VB_START);

   IM->Start = VB_START;
   IM->Count = VB_START;
   IM->LastMaterial = (GLuint) (-1);
   IM->LastPrimitive = VB_START;
   IM->Primitive[VB_START] = ctx->Current.Primitive;
   IM->BeginState = VERT_BEGIN_0;         
   IM->ArrayAndFlags = ~ctx->Array.Flags;
   IM->ArrayIncr = ctx->Array.Vertex.Enabled;
}


/* Preserves size information as well.
 */
void 
fixup_4f( GLfloat data[][4], GLuint flag[], GLuint start, GLuint match )
{
   GLuint i = start;

   for (;;) {
      if ((flag[++i] & match) == 0) {
	 COPY_4FV(data[i], data[i-1]);
	 flag[i] |= (flag[i-1] & match);
	 if (flag[i] & VERT_END_VB) break;
      } 
   }
}



/* Only needed for buffers with eval coords.
 */
void
fixup_3f( float data[][3], GLuint flag[], GLuint start, GLuint match )
{
   GLuint i = start;

   for (;;) {
      if ((flag[++i] & match) == 0) {
	 COPY_3V(data[i], data[i-1]);
	 flag[i] |= match;
	 if (flag[i] & VERT_END_VB) break;
      } 
   }
}


void
fixup_1ui( GLuint *data, GLuint flag[], GLuint start, GLuint match )
{
   GLuint i = start;

   for (;;) {
      if ((flag[++i] & match) == 0) {
	 data[i] = data[i-1];
	 if (flag[i] & VERT_END_VB) break;
      } 
   }
   flag[i] |= match;
}

void
fixup_1ub( GLubyte *data, GLuint flag[], GLuint start, GLuint match )
{
   GLuint i = start;

   for (;;) {
      if ((flag[++i] & match) == 0) {
	 data[i] = data[i-1];
	 if (flag[i] & VERT_END_VB) break;
      } 
   }
   flag[i] |= match;
}


void
fixup_4ub( GLubyte data[][4], GLuint flag[], GLuint start, GLuint match )
{
   GLuint i = start;

   for (;;) {
      if ((flag[++i] & match) == 0) {
	 COPY_4UBV(data[i], data[i-1]);
	 if (flag[i] & VERT_END_VB) break;
      } 
   }
   flag[i] |= match;
}


/* Do this instead of fixup for shared normals.
 */
void
find_last_3f( float data[][3], GLuint flag[], GLuint match, GLuint count )
{
   GLuint i = count;

   for (;;) 
      if ((flag[--i] & match) != 0) {
	 COPY_3V(data[count], data[i]);	 
	 return;      
      }
}

void
fixup_first_4v( GLfloat data[][4], GLuint flag[], GLuint match, 
		GLuint start, GLfloat *dflt )
{
   GLuint i = start-1;
   match |= VERT_END_VB;

   while ((flag[++i]&match) == 0)
      COPY_4FV(data[i], dflt);
}


void
fixup_first_1ui( GLuint data[], GLuint flag[], GLuint match,
		 GLuint start, GLuint dflt )
{
   GLuint i = start-1;
   match |= VERT_END_VB;

   while ((flag[++i]&match) == 0)
      data[i] = dflt;
}


void
fixup_first_1ub( GLubyte data[], GLuint flag[], GLuint match,
		 GLuint start, GLubyte dflt )
{
   GLuint i = start-1;
   match |= VERT_END_VB;

   while ((flag[++i]&match) == 0)
      data[i] = dflt;
}


void
fixup_first_4ub( GLubyte data[][4], GLuint flag[], GLuint match,
		 GLuint start, GLubyte dflt[4] )
{
   GLuint i = start-1;
   match |= VERT_END_VB;

   while ((flag[++i]&match) == 0)
      COPY_4UBV(data[i], dflt);
}




static GLuint vertex_sizes[16] = { 0, 
				   1, 
				   2, 2, 
				   3, 3, 3, 3, 
				   4, 4, 4, 4, 4, 4, 4, 4 };



GLuint gl_texcoord_size( GLuint flag, GLuint unit )
{
   flag >>= VERT_TEX0_SHIFT + unit * NR_TEXSIZE_BITS;
   return vertex_sizes[flag & 0xf];
}


static void set_vec_sizes( struct immediate *IM, GLuint orflag )
{
   GLuint i;

   if (orflag & VERT_OBJ_ANY) {
      GLuint szflag = orflag & VERT_OBJ_234;
      IM->v.Obj.size = vertex_sizes[szflag<<1];
   }

   for (i = 0 ; i < MAX_TEXTURE_UNITS ; i++) {
      if (orflag & VERT_TEX_ANY(i)) {
	 GLuint szflag = ((orflag>>(VERT_TEX0_SHIFT+i*NR_TEXSIZE_BITS))
			  & 0xf);
	 IM->v.TexCoord[i].size = vertex_sizes[szflag];
      }
   }
}


void gl_compute_orflag( struct immediate *IM )
{
   GLuint count = IM->Count;
   GLuint orflag = 0;
   GLuint andflag = ~0;
   GLuint i;

   IM->LastData = count-1;

   /* This was previously done inside the immediate API functions 
    */
   for (i = VB_START ; i < count ; i++) {
      andflag &= IM->Flag[i];
      orflag |= IM->Flag[i];
   }

   if (IM->Flag[i] & VERT_DATA) {
      IM->LastData++;
      andflag &= IM->Flag[i];
      orflag |= IM->Flag[i];
   }      

   IM->Flag[IM->LastData+1] |= VERT_END_VB;
   IM->AndFlag = andflag;
   IM->OrFlag = orflag;
}

void gl_fixup_input( GLcontext *ctx, struct immediate *IM )
{
   GLuint count = IM->Count;
   GLuint start = VB_START;
   GLuint fixup, diff;
   GLuint andflag = IM->AndFlag;
   GLuint orflag = IM->OrFlag;

   IM->Primitive[count] = IM->Primitive[IM->LastPrimitive];
   IM->NextPrimitive[IM->LastPrimitive] = count;
   IM->NextPrimitive[count] = count+1;

   if (MESA_VERBOSE&VERBOSE_IMMEDIATE) {
      fprintf(stderr, "Count: %u LastData: %u\n", IM->Count, IM->LastData);
      gl_print_vert_flags("Orflag", orflag);
      gl_print_vert_flags("Andflag", andflag);
   }

   /* Array elements modify the current state - must do this before
    * fixup.
    */
   if (ctx->CompileCVAFlag && !(andflag & VERT_ELT)) 
      gl_rescue_cva( ctx, IM );

   if (orflag & VERT_ELT) 
   {
      if (!ctx->CompileCVAFlag)
	 gl_exec_array_elements( ctx, IM );

      orflag = IM->OrFlag;
      andflag = IM->AndFlag;
      start = IM->Start;
   }
   
   fixup = ~andflag & VERT_FIXUP;

   if (!ctx->CompileFlag) 
      fixup &= ctx->CVA.elt.inputs; 
   
   if (!ctx->ExecuteFlag)
      fixup &= orflag;

   if (ctx->CompileCVAFlag)
      fixup &= ~ctx->CVA.pre.outputs;

   if ((orflag & (VERT_OBJ_ANY|VERT_EVAL_ANY)) == 0)
      fixup = 0;

   if (fixup) 
   {

      if (ctx->ExecuteFlag && (fixup & ~IM->Flag[VB_START])) {
	 GLuint copy = fixup & ~IM->Flag[VB_START];

	 if (copy & VERT_NORM) 	 
	    COPY_3V( IM->Normal[VB_START], ctx->Current.Normal );
	 
	 if (copy & VERT_RGBA)
	    COPY_4UBV( IM->Color[VB_START], ctx->Current.ByteColor);
      
	 if (copy & VERT_INDEX)
	    IM->Index[VB_START] = ctx->Current.Index;
	 
	 if (copy & VERT_EDGE)
	    IM->EdgeFlag[VB_START] = ctx->Current.EdgeFlag;

	 if (copy & VERT_TEX0_ANY)
	    COPY_4FV( IM->TexCoord[0][VB_START], ctx->Current.Texcoord[0] ); 
	 
	 if (copy & VERT_TEX1_ANY)
	    COPY_4FV( IM->TexCoord[1][VB_START], ctx->Current.Texcoord[1] ); 
      }


      if (MESA_VERBOSE&VERBOSE_IMMEDIATE)
	 gl_print_vert_flags("fixup", fixup);
      
      if (fixup & VERT_TEX0_ANY) {
	 if (orflag & VERT_TEX0_ANY)
	    fixup_4f( IM->TexCoord[0], IM->Flag, start, VERT_TEX0_1234 );
	 else 
	    fixup_first_4v( IM->TexCoord[0], IM->Flag, 0, start, 
			    IM->TexCoord[0][3]);
      }

      if (fixup & VERT_TEX1_ANY) {
	 if (orflag & VERT_TEX1_ANY)
	    fixup_4f( IM->TexCoord[1], IM->Flag, start, VERT_TEX1_1234 );
	 else 
	    fixup_first_4v( IM->TexCoord[1], IM->Flag, 0, start, 
			    IM->TexCoord[1][3] );
      }

      if (fixup & VERT_EDGE) {
	 if (orflag & VERT_EDGE)
	    fixup_1ub( IM->EdgeFlag, IM->Flag, start, VERT_EDGE );
	 else 
	    fixup_first_1ub( IM->EdgeFlag, IM->Flag, 0, start, 
			     IM->EdgeFlag[3] );
      }

      if (fixup & VERT_INDEX) {
	 if (orflag & VERT_INDEX)
	    fixup_1ui( IM->Index, IM->Flag, start, VERT_INDEX );
	 else 
	    fixup_first_1ui( IM->Index, IM->Flag, 0, start, IM->Index[3] );
      }

      if (fixup & VERT_RGBA) {
	 if (orflag & VERT_RGBA) 
	    fixup_4ub( IM->Color, IM->Flag, start, VERT_RGBA );      
	 else 
	    fixup_first_4ub( IM->Color, IM->Flag, 0, start, IM->Color[3] );
      }

      if (fixup & VERT_NORM) {
	 if (IM->OrFlag & VERT_EVAL_ANY)
	    fixup_3f( IM->Normal, IM->Flag, start, VERT_NORM );
	 else if (!(IM->Flag[IM->LastData] & VERT_NORM)) 
 	    find_last_3f( IM->Normal, IM->Flag, VERT_NORM, IM->LastData );
      }
   }

   diff = count - start;
   IM->v.Obj.count = diff;
   IM->v.Normal.count = diff;
   IM->v.TexCoord[0].count = diff;
   IM->v.TexCoord[1].count = diff;
   IM->v.EdgeFlag.count = diff;
   IM->v.Color.count = diff;
   IM->v.Index.count = diff;
 
   /* Prune possible half-filled slot.
    */
   IM->Flag[IM->LastData+1] &= ~VERT_END_VB;
   IM->Flag[IM->Count] |= VERT_END_VB;

/*    gl_print_cassette( IM, 0, ~0 ); */
}


static void calc_normal_lengths( GLfloat *dest, 
				 CONST GLfloat (*data)[3],
				 const GLuint *flags,
				 GLuint count )
{
   GLuint i;

   for (i = VB_START ; i < count ; i++ )
      if (flags[i] & VERT_NORM) {
	 GLfloat tmp = LEN_3FV( data[i] );
	 dest[i] = 0;
	 if (tmp > 0)
	    dest[i] = 1.0 / tmp;
      }      
}


/* Revive a compiled immediate struct - propogate new 'Current'
 * values.  Often this is redundant because the current values were
 * known and fixed up at compile time.
 */
void gl_fixup_cassette( GLcontext *ctx, struct immediate *IM )
{
   GLuint fixup;
   
   if (IM->Count == VB_START) 
      return;

   if (ctx->NewState)
      gl_update_state( ctx );

   fixup = ctx->CVA.elt.inputs & ~IM->AndFlag & VERT_FIXUP;

   if (ctx->Transform.Normalize && !IM->NormalLengths) {
      IM->NormalLengths = (GLfloat *)malloc(sizeof(GLfloat) * (IM->Count+1));
      calc_normal_lengths( IM->NormalLengths, 
			   (const GLfloat (*)[3])IM->Normal, 
			   IM->Flag, IM->Count);
   }

   if (fixup) {

      if (MESA_VERBOSE & VERBOSE_IMMEDIATE)
	 gl_print_vert_flags("fixup_cassette", fixup);

      if (fixup & VERT_TEX0_ANY) 
	 fixup_first_4v( IM->TexCoord[0], IM->Flag, VERT_TEX0_ANY, VB_START,
			 ctx->Current.Texcoord[0] );

      if (fixup & VERT_TEX1_ANY) 
	 fixup_first_4v( IM->TexCoord[1], IM->Flag, VERT_TEX1_ANY, VB_START,
			 ctx->Current.Texcoord[1] );

      if (fixup & VERT_EDGE)
	 fixup_first_1ub(IM->EdgeFlag, IM->Flag, VERT_EDGE, VB_START,
			 ctx->Current.EdgeFlag );

      if (fixup & VERT_INDEX)
	 fixup_first_1ui(IM->Index, IM->Flag, VERT_INDEX, VB_START,
			 ctx->Current.Index );
   
      if (fixup & VERT_RGBA) 
	 fixup_first_4ub(IM->Color, IM->Flag, VERT_RGBA, VB_START,
			 ctx->Current.ByteColor );      

      if ((fixup & VERT_NORM) && !(IM->Flag[VB_START] & VERT_NORM)) {
	 COPY_3V(IM->Normal[VB_START], ctx->Current.Normal);
	 if (ctx->Transform.Normalize)
	    IM->NormalLengths[VB_START] = 1.0 / LEN_3FV(ctx->Current.Normal);
      }
   }  
}



static void fixup_primitives( GLcontext *ctx, struct immediate *IM )
{
   static GLuint increment[GL_POLYGON+2] = { 1,2,1,1,3,1,1,4,2,1,1 };
   static GLuint intro[GL_POLYGON+2]     = { 0,0,2,2,0,2,2,0,2,2,0 };
   struct vertex_buffer *VB = ctx->VB;
   const GLuint *flags = IM->Flag;
   const GLuint *in_prim = IM->Primitive;
   const GLuint *in_nextprim = IM->NextPrimitive;
   GLuint *out_prim = VB->IM->Primitive;
   GLuint *out_nextprim = VB->IM->NextPrimitive;
   GLuint count = VB->Count;
   GLuint in, out, last;
   GLuint incr, prim;   
   GLuint transition;
   GLuint interesting;
   GLuint err;
   
/*     printf("IM: %d prim[3]: %d\n", IM->id, IM->Primitive[3]); */

   if (ctx->Current.Primitive == GL_POLYGON+1) {
      transition = VERT_BEGIN;
      err = IM->BeginState & VERT_ERROR_1;
   } else {
      transition = VERT_END;
      err = IM->BeginState & VERT_ERROR_0;
   }

   if (err) {
      /* Occurred somewhere inside the vb.  Don't know/care where/why.
       */
      gl_error( ctx, GL_INVALID_OPERATION, "begin/end");
   }

   interesting = transition | VERT_END_VB;


   for (in = VB_START ; in <= count ; in = in_nextprim[in]) 
      if (flags[in] & interesting) 
	 break;

   out = VB->CopyStart;

   if (in == out) {
      out_nextprim[out] = in_nextprim[in];
      out_prim[out] = in_prim[in];
      last = IM->LastPrimitive;
   } else if (flags[in] & transition) {
      out_nextprim[out] = in;
      out_prim[out] = ctx->Current.Primitive;
      out = in;
      last = IM->LastPrimitive;
   } else {
      out_nextprim[out] = in;
      out_prim[out] = ctx->Current.Primitive;
      in++;
      last = out;
   }

   for ( ; in <= count ; in = in_nextprim[in] ) {
      out_prim[in] = in_prim[in];
      out_nextprim[in] = in_nextprim[in];
   }
	  

   VB->Primitive = out_prim;
   VB->NextPrimitive = out_nextprim;
   VB->LastPrimitive = last;
   prim = ctx->Current.Primitive = (GLenum) out_prim[last];


   /* Calculate whether the primitive finished on a 'good number' of
    * vertices, or whether there are overflowing vertices we have to
    * copy to the next buffer (in addition to the normal ones required
    * by a continuing primitive).
    *
    * Note that GL_POLYGON+1, ie outside begin/end, has increment 1.
    */

   incr = increment[prim];
  
   if (incr != 1 && (count - last - intro[prim])) 
      VB->Ovf = (count - last - intro[prim]) % incr; 
   else
      VB->Ovf = 0;

   if (0) 
      fprintf(stderr, "prim: %s count %u last %u incr %u ovf: %u\n",
	      gl_prim_name[prim], count, last, incr, VB->Ovf);
}


void gl_copy_to_current( GLcontext *ctx, struct immediate *IM )
{
   GLuint count = IM->LastData;
   GLuint flag = IM->OrFlag;
   GLuint mask = 0;

   if (MESA_VERBOSE&VERBOSE_IMMEDIATE)
      gl_print_vert_flags("copy to current", flag);

   if (flag & VERT_NORM) 
      COPY_3FV( ctx->Current.Normal, IM->Normal[count]);   
   
   if (flag & VERT_INDEX)
      ctx->Current.Index = IM->Index[count];

   if (flag & VERT_EDGE)
      ctx->Current.EdgeFlag = IM->EdgeFlag[count];

   if (flag & VERT_RGBA) 
      COPY_4UBV(ctx->Current.ByteColor, IM->Color[count]);

   if (flag & VERT_TEX0_ANY) {
      mask |= VERT_TEX0_1234;
      COPY_4FV( ctx->Current.Texcoord[0], IM->TexCoord[0][count]);
   }
      
   if (flag & VERT_TEX1_ANY) {
      mask |= VERT_TEX1_1234;
      COPY_4FV( ctx->Current.Texcoord[1], IM->TexCoord[1][count]);
   }

   /* Save the texcoord size information as well.
    */
   ctx->Current.Flag &= ~mask;
   ctx->Current.Flag |= IM->Flag[count] & mask;
}


void gl_execute_cassette( GLcontext *ctx, struct immediate *IM )
{
   struct vertex_buffer *VB = ctx->VB;
   struct immediate *prev = VB->prev_buffer;


   IM->ref_count++;

   if (prev != IM || IM != VB->IM ) {
      gl_copy_prev_vertices( VB, VB->prev_buffer, IM );
   }

   if (! --prev->ref_count ) 
      gl_immediate_free( prev );

   VB->prev_buffer = IM;
   VB->Start = IM->Start;
   VB->Count = IM->Count;
   VB->Flag = IM->Flag;
   VB->AndFlag = IM->AndFlag;
   VB->OrFlag = IM->OrFlag | VB->SavedOrFlag;
   VB->EltPtr = &IM->v.Elt;
   VB->MaterialMask = IM->MaterialMask;
   VB->Material = IM->Material;
   VB->CullMode = (IM->AndFlag & VERT_NORM) ? 0 : COMPACTED_NORMALS;
   VB->ObjPtr = &IM->v.Obj;
   VB->NormalPtr = &IM->v.Normal;
   VB->ColorPtr = &IM->v.Color;
   VB->Color[0] = VB->Color[1] = VB->ColorPtr;
   VB->IndexPtr = &IM->v.Index;
   VB->EdgeFlagPtr = &IM->v.EdgeFlag;
   VB->TexCoordPtr[0] = &IM->v.TexCoord[0];
   VB->TexCoordPtr[1] = &IM->v.TexCoord[1];
   VB->BoundsPtr = IM->Bounds;
   VB->NormalLengthPtr = IM->NormalLengths;
   VB->IndirectCount = VB->Count;
   VB->SavedOrFlag = 0;

/*     ctx->CVA.orflag |= VB->OrFlag; */


   gl_copy_to_current( ctx, IM );

   set_vec_sizes( IM, VB->OrFlag );

   if (MESA_VERBOSE&VERBOSE_IMMEDIATE) 
      fprintf(stderr,
	      "executing cassette, rows %u tc0->size == %u tc1->size == %u\n",
	      VB->Count,
	      VB->TexCoordPtr[0]->size,
	      VB->TexCoordPtr[1]->size);

   if (MESA_VERBOSE&VERBOSE_IMMEDIATE) 
      gl_print_cassette( IM, 1, ~0 );

   
   if (IM->OrFlag & VERT_EVAL_ANY) 
      gl_eval_vb( VB );

   if (IM->Count > VB_START || 
       (IM->Flag[VB_START] & (VERT_END|VERT_BEGIN)))
      fixup_primitives( ctx, IM );

   if (VB->IndirectCount > VB_START) 
      gl_run_pipeline( VB );      
   else 
      gl_update_materials( VB );

   gl_reset_vb( VB );
}



void gl_print_cassette( struct immediate *IM, GLboolean fixed_up, GLuint req ) 
{
   GLuint i;
   GLuint andflag = IM->AndFlag;
   GLuint orflag = IM->OrFlag;
   GLuint state = IM->BeginState;
   GLuint *flags = IM->Flag;
   static const char *tplate[5] = { "%s ", 
				    "%s: %f ",
				    "%s: %f %f ",
				    "%s: %f %f %f ",
				    "%s: %f %f %f %f " };
   (void) fixed_up;

   fprintf(stderr, "Cassette, %u rows.\n", IM->Count);

   gl_print_vert_flags("Contains at least one", orflag);

   if (IM->Count != VB_START) 
   {
      gl_print_vert_flags("Contains a full complement of", andflag);
   
      fprintf(stderr, "Final begin/end state %s/%s, errors %s/%s\n",
	     (state & VERT_BEGIN_0) ? "in" : "out", 
	     (state & VERT_BEGIN_1) ? "in" : "out", 
	     (state & VERT_ERROR_0) ? "y" : "n", 
	     (state & VERT_ERROR_1) ? "y" : "n");

      fprintf(stderr, "Obj size: %u, TexCoord0 size: %u, TexCoord1 size: %u\n",
	     IM->v.Obj.size,
	     IM->v.TexCoord[0].size,
	     IM->v.TexCoord[1].size);
   }

   if (!req) return;

   for (i = 0 ; i <= IM->Count ; i++) {
      fprintf(stderr, "%u: ", i);
      if (req & VERT_OBJ_ANY) {
	 if (flags[i] & VERT_EVAL_C1) 
	    fprintf(stderr, "EvalCoord %f ", IM->Obj[i][0]);
	 else if (flags[i] & VERT_EVAL_P1) 
	    fprintf(stderr, "EvalPoint %.0f ", IM->Obj[i][0]);
	 else if (flags[i] & VERT_EVAL_C2) 
	    fprintf(stderr, "EvalCoord %f %f ", IM->Obj[i][0], IM->Obj[i][1]);
	 else if (flags[i] & VERT_EVAL_P2) 
	    fprintf(stderr, "EvalPoint %.0f %.0f ", IM->Obj[i][0], IM->Obj[i][1]);
	 else if ((flags[i] & (VERT_ELT|VERT_OBJ_ANY)) == (VERT_ELT)) 
	    fprintf(stderr, "Array elt %u\t", IM->Elt[i]);
	 else if (i < IM->Count && (flags[i]&VERT_OBJ_234)) {
	    fprintf(stderr, "(%x) ", flags[i] & VERT_OBJ_234);
	    fprintf(stderr, tplate[vertex_sizes[(flags[i]&VERT_OBJ_234)<<1]],
		   "Obj",
		   IM->Obj[i][0], IM->Obj[i][1], IM->Obj[i][2], IM->Obj[i][3]);
	 }
      }

      if (req & flags[i] & VERT_NORM)
	 fprintf(stderr, " Norm %f %f %f ", 
		IM->Normal[i][0], IM->Normal[i][1], IM->Normal[i][2]);

      if (req & flags[i] & VERT_TEX0_ANY)
	 fprintf(stderr, tplate[vertex_sizes[(flags[i]>>VERT_TEX0_SHIFT)&7]], 
		"TC0",
		IM->TexCoord[0][i][0], IM->TexCoord[0][i][1], 
		IM->TexCoord[0][i][2], IM->TexCoord[0][i][2]);


      if (req & flags[i] & VERT_TEX1_ANY)
	 fprintf(stderr, tplate[vertex_sizes[(flags[i]>>(VERT_TEX0_SHIFT+4))&7]], 
		"TC1",
		IM->TexCoord[1][i][0], IM->TexCoord[1][i][1], 
		IM->TexCoord[1][i][2], IM->TexCoord[1][i][2]);

      if (req & flags[i] & VERT_RGBA)
	 fprintf(stderr, " Rgba %d %d %d %d ", 
		IM->Color[i][0], IM->Color[i][1], 
		IM->Color[i][2], IM->Color[i][3]);

      if (req & flags[i] & VERT_INDEX)
	 fprintf(stderr, " Index %u ", IM->Index[i]);
      
      if (req & flags[i] & VERT_EDGE)
	 fprintf(stderr, " Edgeflag %d ", IM->EdgeFlag[i]);

      /* The order of these two is not easily knowable, but this is
       * the usually correct way to look at them.
       */
      if (req & flags[i] & VERT_END) 
	 fprintf(stderr, " END ");

      if (req & flags[i] & VERT_BEGIN) 
	 fprintf(stderr, " BEGIN(%s) ", gl_prim_name[IM->Primitive[i]]);

      fprintf(stderr, "\n");  
   }      
}













