/* $Id: vertices.c,v 1.3 1999/09/10 16:32:02 keithw Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <stdio.h>
#include "types.h"
#include "vertices.h"

#if defined(USE_X86_ASM)
#include "X86/common_x86asm.h"
#endif

/* The start of a bunch of vertex oriented geometry routines.  These
 * are expected to support the production of driver-specific fast paths
 * for CVA and eventually normal processing.
 *
 * These have been taken from fxfastpath.c, and are now also used in
 * the mga driver.
 *
 * These should grow to include:
 *     - choice of 8/16 dword vertices
 *     - ??
 *     - use of portable assembly layouts.
 *
 * More tentatively:
 *     - more (all?) matrix types
 *     - more (?) vertex sizes
 *
 * -- Keith Whitwell.
 */

/* The inline 3dnow code seems to give problems with some peoples
 * compiler/binutils. 
 */
#undef USE_3DNOW_ASM


#if defined(USE_X86_ASM) && defined(__GNUC__)

#if defined(USE_3DNOW_ASM)
/* KW: A single loop seems to be faster than two loops for large
 * stride output vertices.  My guess is that cache misses are the
 * culprit, but it might just be my coding. 
 */
static void gl_3dnow_transform_v16 (GLfloat *dest,
				    const GLvector4f *vec,
				    const GLfloat *m,
				    GLuint count )
{
   int stride = vec->stride;
   float *from = vec->start;
   int c = count;

   int d0, d1, d2, d3, d4;

__asm__ __volatile__ (

"		femms   \n"
"  		test		%3,%3   \n"
"		jle		3f   \n"
"   \n"
"		movq		48(%1),%%mm7  # m31         | m30     \n"
"		movq		56(%1),%%mm3  # m33         | m32   \n"
"   \n"
".balign	32   \n"
"2:   \n"
"               prefetchw       128(%0)       # prefetch 2 output verts ahead\n"
"		movq		(%2),  %%mm0  # x1          | x0   \n"
"		movd		8(%2), %%mm1  #             | x2   \n"
"		movq		(%1),  %%mm4  # m00         | m01   \n"
"		punpckhdq	%%mm0, %%mm2  # x1          |       \n"
"		movq		16(%1),%%mm5  # m10         | m11   \n"
"		punpckldq	%%mm0, %%mm0  # x0          | x0   \n"
"		movq		32(%1),%%mm6  # m20         | m21   \n"
" 		pfmul		%%mm0, %%mm4  # x0.m01      | x0.m00   \n"
"		punpckhdq	%%mm2, %%mm2  # x1          | x1   \n"
"		pfmul		%%mm2, %%mm5  # x1.m11      | x1.m10   \n"
"		punpckldq	%%mm1, %%mm1  # x2          | x2   \n"
"		pfmul		8(%1), %%mm0  # x0.m03      | x0.m02   \n"
"		pfmul		24(%1),%%mm2  # x1.m13      | x1.m12   \n"
"		pfmul		%%mm1, %%mm6  # x2.m21      | x2.m20   \n"
"		pfadd		%%mm4, %%mm5  # x0m01+x1m11 | x0m00+x1m10   \n"
"		pfmul		40(%1),%%mm1  # x2.m21      | x2.m20   \n"
"		pfadd		%%mm0, %%mm2  # x0m03+x1m13 | x0m02+x1m12   \n"
"		pfadd		%%mm5, %%mm6  # (xm012)1    | (xm012)0   \n"
"		pfadd		%%mm1, %%mm2  # (xm012)3    | (xm012)2   \n"
"		pfadd		%%mm7, %%mm6  # r1          | r0   \n"
"		pfadd		%%mm3, %%mm2  # r3          | r2   \n"
"		add		%4,    %2     # next input     \n"
"		movq		%%mm6, (%0)   #    \n"
"		movq		%%mm2, 8(%0)  #   \n"
"		add		$64,   %0     # next output   \n"
"		dec		%3   \n"
"		jne		2b   \n"
"   \n"
"3:   \n"
"		femms   \n"

/* eax: dest
 * esi: mat (was ebx)
 * edx: from
 * ecx: count
 * edi: input stride
 */

	 : "=&a"(d0), "=&S"(d1), "=&d"(d2), "=&c"(d3), "=&D"(d4)
	 : "0" (dest), "1" (m), "2" (from), "3" (c), "4" (stride) 
	 : "memory"); 
}

#endif


/* This is nothing more glamorous than an objdump of one of Josh's
 * routines hacked to match the above.    
 */
static void gl_x86_transform_v16 (GLfloat *dest,
				  const GLvector4f *vec,
				  const GLfloat *m,
				  GLuint count )
{
   int stride = vec->stride;
   float *from = vec->start;

   int d0, d1, d2, d3, d4;

__asm__ __volatile__ (

"             testl  %3, %3    \n"
"             je     3f    \n"
".balign 16    \n"
"2:    \n"
"             flds   0x0(%2)    \n"
"             fmuls  0x0(%1)    \n"
"             flds   0x0(%2)    \n"
"             fmuls  0x4(%1)    \n"
"             flds   0x0(%2)    \n"
"             fmuls  0x8(%1)    \n"
"             flds   0x0(%2)    \n"
"             fmuls  0xc(%1)    \n"
"             flds   0x4(%2)    \n"
"             fmuls  0x10(%1)    \n"
"             flds   0x4(%2)    \n"
"             fmuls  0x14(%1)    \n"
"             flds   0x4(%2)    \n"
"             fmuls  0x18(%1)    \n"
"             flds   0x4(%2)    \n"
"             fmuls  0x1c(%1)    \n"
"             fxch   %%st(3)    \n"
"             faddp  %%st,%%st(7)    \n"
"             fxch   %%st(1)    \n"
"             faddp  %%st,%%st(5)    \n"
"             faddp  %%st,%%st(3)    \n"
"             faddp  %%st,%%st(1)    \n"
"             flds   0x8(%2)    \n"
"             fmuls  0x20(%1)    \n"
"             flds   0x8(%2)    \n"
"             fmuls  0x24(%1)    \n"
"             flds   0x8(%2)    \n"
"             fmuls  0x28(%1)    \n"
"             flds   0x8(%2)    \n"
"             fmuls  0x2c(%1)    \n"
"             fxch   %%st(3)    \n"
"             faddp  %%st,%%st(7)    \n"
"             fxch   %%st(1)    \n"
"             faddp  %%st,%%st(5)    \n"
"             faddp  %%st,%%st(3)    \n"
"             faddp  %%st,%%st(1)    \n"
"             fxch   %%st(3)    \n"
"             fadds  0x30(%1)    \n"
"             fxch   %%st(2)    \n"
"             fadds  0x34(%1)    \n"
"             fxch   %%st(1)    \n"
"             fadds  0x38(%1)    \n"
"             fxch   %%st(3)    \n"
"             fadds  0x3c(%1)    \n"
"             fxch   %%st(2)    \n"
"             fstps  0x0(%0)    \n"
"             fstps  0x4(%0)    \n"
"             fxch   %%st(1)    \n"
"             fstps  0x8(%0)    \n"
"             fstps  0xc(%0)    \n"
"             addl   $64,%0    \n"
"             addl   %4,%2    \n"
"             decl   %3       \n"
"             jne    2b    \n"
"3:    \n"

/* eax: dest
 * esi: mat (was ebx, egcs barfed)
 * edx: from
 * ecx: count
 * edi: input stride
 */

	 : "=&a"(d0), "=&S"(d1), "=&d"(d2), "=&c"(d3), "=&D"(d4)
	 : "0" (dest), "1" (m), "2" (from), "3" (count), "4" (stride) 
	 : "memory"); 
}

#endif


static void transform_v16(GLfloat *f,
			  const GLvector4f *vec,
			  const GLfloat *m,
			  GLuint count )
{
   GLfloat *obj =  vec->start;
   GLuint obj_stride = vec->stride;
   GLuint i;

   for (i = 0 ; i < count ; i++, STRIDE_F(obj, obj_stride), f+=16) 
   {
      const GLfloat ox = obj[0], oy = obj[1], oz = obj[2];
      f[0] = m[0] * ox + m[4] * oy + m[8]  * oz + m[12];
      f[1] = m[1] * ox + m[5] * oy + m[9]  * oz + m[13];
      f[2] = m[2] * ox + m[6] * oy + m[10] * oz + m[14];
      f[3] = m[3] * ox + m[7] * oy + m[11] * oz + m[15];
   }
}


static void cliptest_v16( GLfloat *first,
			  GLfloat *last,
			  GLubyte *p_clipOr,
			  GLubyte *p_clipAnd,
			  GLubyte *clipmask )
{	
   GLubyte clipAnd = ~0;
   GLubyte clipOr = 0;
   GLfloat *f = first;
   static int i;
   i = 0;

   for ( ; f != last ; f+=16, clipmask++, i++) 
   {
      const GLfloat cx = f[0];
      const GLfloat cy = f[1];
      const GLfloat cz = f[2];
      const GLfloat cw = f[3];      
      GLubyte mask = 0;

      if (cx >  cw) mask |= CLIP_RIGHT_BIT;
      if (cx < -cw) mask |= CLIP_LEFT_BIT;
      if (cy >  cw) mask |= CLIP_TOP_BIT;
      if (cy < -cw) mask |= CLIP_BOTTOM_BIT;
      if (cz >  cw) mask |= CLIP_FAR_BIT;
      if (cz < -cw) mask |= CLIP_NEAR_BIT;

      *clipmask = mask;
      clipAnd &= mask;
      clipOr |= mask;
   }

   (*p_clipOr) |= clipOr;
   (*p_clipAnd) &= clipAnd;
}



GLenum gl_reduce_prim[GL_POLYGON+1] = {
   GL_POINTS,
   GL_LINES,
   GL_LINES,
   GL_LINES,
   GL_TRIANGLES,
   GL_TRIANGLES,
   GL_TRIANGLES,
   GL_TRIANGLES,
   GL_TRIANGLES,
   GL_TRIANGLES,
};  

gl_transform_func gl_transform_points3_vertex16_general = 0;
gl_cliptest_func gl_cliptest_points4_vertex16;			      

void gl_init_vertices()
{
   gl_cliptest_points4_vertex16 = cliptest_v16;
   gl_transform_points3_vertex16_general = transform_v16;

#if defined(USE_X86_ASM) && defined(__GNUC__)
   if (gl_x86_cpu_features)
   { 
      extern void gl_x86_FX_cliptest_points4(GLfloat *first_vert,
					     GLfloat *last_vert,
					     GLubyte *or_mask,
					     GLubyte *and_mask,
					     GLubyte *clip_mask );

      
      gl_cliptest_points4_vertex16 = gl_x86_FX_cliptest_points4;
      gl_transform_points3_vertex16_general = gl_x86_transform_v16;

#ifdef USE_3DNOW_ASM     
      if (gl_x86_cpu_features & GL_CPU_3Dnow) {
	 gl_transform_points3_vertex16_general = gl_3dnow_transform_v16;
      }
#endif
   }
#endif
}
